------------------------------------------------------------
-- API Documentation.lua
--
-- This file only exists as an API documentation for module developers, it is not included in
-- CompactRaid's TOC and will not be loaded by the game's LUA script interpreter. Any changes
-- made in this file will by no means affect CompactRaid in any matter.

-- Abin
-- 2012/1/03 Updated to v2.0
------------------------------------------------------------

------------------------------------------------------------
-- How do I create a working module for CompactRaid?
------------------------------------------------------------

-- First you need to register a module to CompactRaid by issuing a call to CompactRaid:RegisterModule, prototype of which shows below:

------------------------------------------------------------
-- CompactRaid:RegisterModule("addonName", "name" [, "title" [, "version" [, "description" [, dualTalents [, parent]]]]])
------------------------------------------------------------

-- Parameters:
-- "addonName" (string), required, the addon name to which the module belongs, can be obtained by add this line to top of your module addon: local addonName = ...
-- "name" (string), required, must be unique among all modules
-- "title" (string), optional, if presents, an option page will be created for your module and this string will be used as the page title
-- "version" (string), optional, the option page displays this string as a version
-- "description" (string), optional, a brief introduction of your module to be displayed on the option page
-- dualTalents (boolean), whether this module supports dual talents, if true, the module will be notified when the player switches talent groups
-- parent (table), object of the parent module, if presents, module option page will be positioned as a sub-category of "parent"

local module = CompactRaid:RegisterModule("MyFirstAddon", "MyFirstModule", "My First Module", "1.0", "Hello, this is my first module for CompactRaid!")
if not module then return end -- For safety. It only fails when the name "MyFirstModule" is already taken by another module

-- At this point the "module" object is alive and registered to CompactRaid, it has been assigned with a number of function, as the following:

------------------------------------------------------------
-- module:Print("message" [, r, g, b]) -- Prints a message to screen
-- module:RequestRestoreDefaults(["text" [, arg1]]) -- Request a "restore-defaults" operation, details described later.
-- module:RegisterEvent("event" [, "handler"]) -- Register an event like a frame...
-- module:UnregisterEvent("event")
-- module:IsEventRegistered("event")
-- module:RegisterAllEvents()
-- module:UnregisterAllEvents()
------------------------------------------------------------

-- Now you should define at least some of the following callback functions for your module to make it useful.

------------------------------------------------------------
-- module:OnInitialize
------------------------------------------------------------
function module:OnInitialize(db, chardb, firstTime)
	-- Called only once per game session when all saved user data are loaded, no other members of your module will be called
	-- before "OnInitialize". This is where you should do one-time-initialization for your module.
	-- Parameters:
	-- db (table), stores cross-character data. This table is also referenced by "module.db"
	-- chardb (table), stores character specialized data. This table is also referenced by "module.chardb"
	-- firstTime (1/nil), whether it's the first time this character uses this module
end

------------------------------------------------------------
-- module:OnTalentGroupChange (new in v1.1)
------------------------------------------------------------
function module:OnTalentGroupChange(talentGroup, talentdb, firstTime)
	-- Called after the player switched his talent groups, also called right after module:OnInitialize once even if the player
	-- does not have dual talent specs available. You will need to switch your data sets accordingly since your module had already
	-- declared as dual talents supported.
	-- Parameters:
	-- talentGroup (number), either 1 or 2, represents the currently activated talent group
	-- talentdb (table), stores talent group specialized data. This table is also referenced by "module.talentdb"
	-- firstTime (1/nil), whether it's the first time this character activates this talent group
end

------------------------------------------------------------
-- module:OnRestoreDefaults
------------------------------------------------------------
function module:OnRestoreDefaults(arg1)
	-- Called when your module requested a "restore-defaults" operation and the player clicked the "OK" button
	-- on the confirmation dialog. At this point you should actually restore and apply default settings to
	-- everything relevant.
	-- Parameters: arg1 (anything), the arbitrary argument you passed to module:RequestRestoreDefaults.

	-- Your module requests a "restore-defaults" operation by a previous call of:
	-- module:RequestRestoreDefaults(["text" [, arg1]])
	-- Where "text" is the specific text to be displayed on the popup confirmation dialog, if absent, a default
	-- text is displayed instead. "arg1" is an arbitrary data of any type which will be passed back to
	-- module:OnRestoreDefaults callback after the player clicks the "OK" button on the confirmation dialog.
end

------------------------------------------------------------
-- module:OnCreate
------------------------------------------------------------
function module:OnCreate(overlay)
	-- Called when a new CompactRaid unit frame is created. "overlay" is not the unit frame itself, instead,
	-- it's an overlay frame created on top of the unit frame and covers which entirely, while you can reach the
	-- underlying unit frame through overlay:GetParent(), it is strongly suggested that you don't. Unit frames are secure
	-- and protected, they shall never be touched unless you have a good reason plus a strong knowledge on dealing with
	-- secure frame templates and such things, taints will likely arise otherwise.

	-- The "overlay" frame is fully anchored to the unit frame and its frame level is higher than any thing else on the
	-- unit frame, you should always place your own UI elements on "overlay" or its children to avoid being obstructed.
end

------------------------------------------------------------
-- module:OnUnitChange
------------------------------------------------------------
function module:OnUnitChange(overlay, unit, realUnit, inVehicle)
	-- Called when the "unit" attribute of the underlying unit frame has changed. If your module cares about unit, you
	-- need to do something here. "unit" is the current unit-id ("raid1", "raid2", etc), "realUnit" differs from "unit"
	-- only when the player is in a vehicle. "inVehicle" (1/nil) indicates whether the unit has a vehicle UI at the moment,
end

------------------------------------------------------------
-- module:OnRangeChange
------------------------------------------------------------
function module:OnRangeChange(overlay, inRange)
	-- Called when the unit referred by the underlying unit frame has just moved in or out of the player's range.
	-- While there's never been an event in WOW that reflects units' in-range status, this callback offers at least a
	-- semi-okay solution. "inRange" (1/nil) indicates whether the unit is in range or not.
end

------------------------------------------------------------
-- module:OnDeathFlagChange
------------------------------------------------------------
function module:OnDeathFlagChange(overlay, flag)
	-- Called when the unit referred by the underlying unit frame has just changed its death status. This is not a very
	-- useful function but, since I already implemented death status checking in main module, why not simply pass them
	-- through in case someone might need it.
	-- "flag" (string), the value of this parameter can be "dead", "ghost", or nil. Where "dead" means the unit is dead
	-- but has not released its spirit, "ghost" means the unit released its spirit and is staying as a ghost, nil indicates
	-- the unit is alive as normal.
end

------------------------------------------------------------
-- module:OnShow
------------------------------------------------------------
function module:OnShow(overlay)
	-- Called when the underlying unit frame shows up
end

------------------------------------------------------------
-- module:OnHide
------------------------------------------------------------
function module:OnHide(overlay)
	-- Called when the underlying unit frame hides
end

------------------------------------------------------------
-- module:OnEnter
------------------------------------------------------------
function module:OnEnter(overlay)
	-- Called when mouse cursor hovers over an unit frame
end

------------------------------------------------------------
-- module:OnLeave
------------------------------------------------------------
function module:OnLeave(overlay)
	-- Called when mouse cursor leaves an unit frame
end


------------------------------------------------------------
-- Aside from above module functions, there also are CompactRaid functions ready for
-- you, the following is a list of those are most commonly used:
------------------------------------------------------------

-- CompactRaid:CheckVersion(version [, "errorTitle"]) -- Check CompactRaid version, the function returns 1 if the current CompactRaid version is euquivalent or greater than "version", or nil otherwise, in which case an error message pops up if "errorTitle" is specified.
-- CompactRaid:GetModule("name") -- Returns the registered module whose name is "name", returns nil if the module dosen't exist
-- CompactRaid:RegisterLocale("name", "locale", data) -- Register a locale for a module
-- CompactRaid:GetLocale("name") -- Retrieves locale data for a module
-- CompactRaid:EnumUnitFrames(...) -- Enumerate all created overlay frames, your callback function returning non-nil value will abort the process
-- CompactRaid:CombatSafeCall("func", ...) -- Call directly and return nothing if not in combat, or pending the call until leaves combat and return the pending handle
-- CompactRaid:CombatSafeCall(func, ...)
-- CompactRaid:CombatSafeCall(object, "method", ...)
-- CompactRaid:CancelPendingCombatSafeCall(handle) -- Cancel a pending call if in combat, "handle" is the pending handle returnd by a previous CompactRaid:CombatSafeCall

------------------------------------------------------------
-- About CompactRaid:EnumFrames(...), the way how this function works depends on arguments list you pass:

-- Arguments			-- Results (the overlay frame is always passed to your callback function)
----------------------------------------------------------------
-- func, ...			-- func(overlay, ...)		-- func is a valid function
-- object, "method", ...	-- object:method(overlay, ...)	-- object is a table, "method" is a member function name of object
-- nil, "method", ...		-- overlay:method(...)		-- first argument is nil, "method" is a member function name of the overlay frame, such as "SetScale", "Show", "Hide", etc
----------------------------------------------------------------

----------------------------------------------------------------
-- Creating unit frames
----------------------------------------------------------------

-- Developers can create their own unit frames and register to CompactRaid. To create an unit frame, you will need to use
-- "CompactRaid:CreateUnitFrame(...)" which maintains all internal work for your frames. The following code snippet creates
-- an unit frame which shows up on center of screen when you target something. The frame gets updated when "PLAYER_TARGET_CHANGED" fires.

local frame = CompactRaid:CreateUnitFrame("MyUnitFrame", UIParent, "target")
frame:SetPoint("CENTER") -- That's it!

----------------------------------------------------------------
-- API's for CompactRaid:CreateUnitFrame(...)
----------------------------------------------------------------

-- frame = CompactRaid:CreateUnitFrame("name", parent, "unit" [, "visibility" [, mini [, "additionalTemplates"]]])

-- "name" - String, frame name, cannot be nil.
-- parent - Frame, specifies parent of the frame.
-- "unit" - String, any valid Blizzard unit-id: "player", "pet", "party1", "raid4", "targettarget", etc.
-- mini - 1/nil, CompactUnitFrame frame style, which determines how the frame appears and how it is resized.
-- "visibility" - String/nil/false, specifies how the frame is shown/hidden, for string it is registered by RegisterStateDriver with "visibility" state,
--                for "false" the developer will maintain its state by himself, for anything else it uses RegisterUnitWatch