﻿------------------------------------------------------------
-- RaidDebuff.lua
--
-- Abin
-- 2010/10/16
------------------------------------------------------------
-- API
--
-- module = CompactRaid:GetModule("RaidDebuff") -- Get the "RaidDebuff" module handle
-- instance = module:RegisterInstance("instance" [, user]) -- Register an instance, user(1/nil) indicates whether this instance can be deleted
-- boss = instance:RegisterBoss(["boss" [, user]]) -- Register a boss, boss name defaults to "module.COMMON_MOBS" if absent
-- success = instance:UnregisterBoss("boss") -- Unregister a boss, user(1/nil) indicates whether this boss can be deleted
-- boss = instance:GetBoss("boss")
-- status = instance:GetUserStatus() -- Get user status of an instance, nil-builtin, 1-builtin modified, 2-user created
-- debuff = boss:RegisterDebuff(spellId [, level [, user]]) -- Register a debuff
-- success = boss:UnregisterDebuff(spellId) -- Unregister a debuff
-- status = boss:GetUserStatus() -- Get user status of a boss, nil-builtin, 1-builtin modified, 2-user created

------------------------------------------------------------

local pairs = pairs
local ipairs = ipairs
local type = type
local floor = floor
local GetSpellInfo = GetSpellInfo
local tinsert = tinsert
local select = select
local GetTime = GetTime
local UnitDebuff = UnitDebuff
local CreateFrame = CreateFrame
local GetRealZoneText = GetRealZoneText
local format = format
local DebuffTypeColor = DebuffTypeColor

local addonName = ...
local L = CompactRaid:GetLocale("RaidDebuff")
local module = CompactRaid:RegisterModule(addonName, "RaidDebuff", L["raid debuff"], "2.0", L["raid debuff desc"])
if not module then return end

module.COMMON_MOBS = GENERAL
module.DEBUFF_LEVEL_NONE = 0
module.DEBUFF_LEVEL_LOW = 1
module.DEBUFF_LEVEL_NORMAL = 2
module.DEBUFF_LEVEL_HIGH = 3
module.instanceList = {}

local activeDebuffs

local function tfind(t, key, value)
	if key and value then
		local i, data
		for i, data in ipairs(t) do
			if data[key] == value then
				return data, i
			end
		end
	end
end

local function NormalizeLevel(level)
	if type(level) == "number" then
		level = floor(level + 0.5)
		if level < module.DEBUFF_LEVEL_NONE then
			level = module.DEBUFF_LEVEL_NONE
		elseif level > module.DEBUFF_LEVEL_HIGH then
			level = module.DEBUFF_LEVEL_HIGH
		end
	else
		level = module.DEBUFF_LEVEL_NORMAL
	end
	return level
end

local function CreateUserData(instance, ...)
	if type(module.db) == "table" and type(module.db.userData) == "table" and type(instance) == "string" then
		local instanceDB = module.db.userData[instance]
		if type(instanceDB) ~= "table" then
			instanceDB = {}
			module.db.userData[instance] = instanceDB
		end

		if type(instanceDB.bosses) ~= "table" then
			instanceDB.bosses = {}
		end

		local args = select("#", ...)
		if args == 0 then
			return instanceDB
		end

		local boss, spellId, level = ...
		if boss == nil then
			boss = module.COMMON_MOBS
		end

		if type(boss) ~= "string" then
			return instanceDB, nil
		end

		local bossDB
		bossDB = instanceDB.bosses[boss]
		if type(bossDB) ~= "table" then
			bossDB = {}
			instanceDB.bosses[boss] = bossDB
		end

		if args > 1 and type(spellId) == "number" then
			bossDB[spellId] = NormalizeLevel(level)
		end

		return instanceDB, bossDB
	end
end

local function DeleteUserData(instance, ...)
	if type(module.db) == "table" and type(module.db.userData) == "table" and type(instance) == "string" then
		local db = module.db.userData
		if not db[instance] then
			return
		end

		local args = select("#", ...)
		if args == 0 then
			db[instance] = nil
			return
		end

		local boss, spellId = ...
		if boss == nil then
			boss = module.COMMON_MOBS
		end

		local bosses = db[instance].bosses
		if not bosses[boss] then
			return
		end

		if args == 1 then
			bosses[boss] = nil
		elseif bosses[boss][spellId] then
			bosses[boss][spellId] = nil
		end
	end
end

local function Boss_GetUserStatus(self)
	if self.user then
		return 2
	end

	local data
	for _, data in pairs(self.debuffs) do
		if data.user and data.boss == self.name then
			return 1
		end
	end
end

local function Boss_RegisterDebuff(self, spellID, level, user)
	if type(spellID) ~= "number" then
		return
	end

	local name, _, icon = GetSpellInfo(spellID)
	if not name then
		return
	end

	level = NormalizeLevel(level)
	local data = { user = user, id = spellID, name = name, icon = icon, link = "|cff71d5ff|Hspell:"..spellID.."|h["..name.."]|h|r", level = level, defLevel = level, boss = self.name }
	self.debuffs[name] = data

	if user then
		CreateUserData(self.instance, self.name, spellID, level)
	end

	return data
end

local function Boss_UnregisterDebuff(self, spellID)
	if type(spellID) == "number" then
		local k, v
		for k, v in pairs(self.debuffs) do
			if v.id == spellID then
				self.debuffs[k] = nil
				if v.user then
					DeleteUserData(self.instance, self.name, spellID)
				end

				return 1
			end
		end
	end
end

local function Instance_GetUserStatus(self)
	if self.user then
		return 2
	end

	local boss
	for _, boss in pairs(self.bosses) do
		if Boss_GetUserStatus(boss) then
			return 1
		end
	end
end

local function Instance_GetBoss(self, boss)
	if not boss or boss == "" then
		boss = module.COMMON_MOBS
	end
	return tfind(self.bosses, "name", boss)
end

local function Instance_RegisterBoss(self, boss, user)
	if not boss or boss == "" then
		boss = module.COMMON_MOBS
	end

	local data, id = Instance_GetBoss(self, boss)
	if data then
		return data, id
	end

	data = { user = user, name = boss, instance = self.name, debuffs = self.debuffs, RegisterDebuff = Boss_RegisterDebuff, UnregisterDebuff = Boss_UnregisterDebuff, GetUserStatus = Boss_GetUserStatus }
	tinsert(self.bosses, data)

	if user then
		CreateUserData(self.name, boss)
	end

	return data, id, 1
end

local function Instance_UnregisterBoss(self, boss)
	local data, id = Instance_GetBoss(self, boss)
	if data then
		tremove(self.bosses, id)
		local k, v
		for k, v in pairs(self.debuffs) do
			if v.boss == boss then
				self.debuffs[k] = nil
			end
		end

		if data.user then
			DeleteUserData(self.name, boss)
		end

		return 1
	end
end

function module:GetInstance(instance)
	return tfind(self.instanceList, "name", instance)
end

function module:RegisterInstance(instance, user)
	if type(instance) ~= "string" then
		return
	end

	local data, id = self:GetInstance(instance)
	if data then
		return data, id
	end

	data = { user = user, RegisterBoss = Instance_RegisterBoss, UnregisterBoss = Instance_UnregisterBoss, RegisterFriendlyBoss = Instance_RegisterBoss, GetBoss = Instance_GetBoss, GetUserStatus = Instance_GetUserStatus, name = instance, bosses = {}, debuffs = {} }
	tinsert(self.instanceList, data)
	id = #self.instanceList

	if user then
		CreateUserData(instance)
	end

	return data, id, 1
end

function module:UnregisterInstance(instance)
	local data, id = self:GetInstance(instance)
	if data then
		tremove(self.instanceList, id)
		if data.user then
			DeleteUserData(instance)
		end
		return 1
	end
end

-----------------------------------------------------------
-- Debuff frame UI
-----------------------------------------------------------

local function FindTopDebuff(unit)
	if not unit or not activeDebuffs then
		return
	end

	local i, data, currentLevel, currentIcon, currentCount, currentDispelType, currentExpires
	for i = 1, 20 do
		local name, _, icon, count, dispelType, _, expires = UnitDebuff(unit, i)
		if not name then
			break
		end

		data = activeDebuffs[name]
		if data and data.level > (currentLevel or 0) then
			currentLevel, currentIcon, currentCount, currentDispelType, currentExpires = data.level, icon, count, dispelType, expires
		end
	end

	return currentLevel, currentIcon, currentCount, currentDispelType, currentExpires
end

local function RaidDebuffFrame_OnUpdate(self, elapsed)
	self.elapsed = (self.elapsed or 0) + elapsed
	if self.elapsed > 0.2 then
		self.elapsed = 0
		local timeLeft = (self.expires or 0) - GetTime()
		if timeLeft > 0 and timeLeft < 15 then
			self.timeText:SetFormattedText("%d", timeLeft)
		else
			self.timeText:SetText()
		end
	end
end

local function RaidDebuffFrame_UpdateDebuff(self)
	local level, icon, count, dispelType, expires = FindTopDebuff(self.unit)
	if level then
		local color
		if dispelType and dispelType ~= "none" then
			color = DebuffTypeColor[dispelType]
		end

		if color then
			self.bkgnd:SetTexture(color.r, color.g, color.b, 1)
		else
			self.bkgnd:SetTexture(0, 0, 0, 0.75)
		end

		self.icon:SetTexture(icon)
		if count and count > 1 then
			self.countText:SetFormattedText("%d", count)
		else
			self.countText:SetText()
		end

		self.expires = expires
		if expires and expires > 0 then
			self:SetScript("OnUpdate", RaidDebuffFrame_OnUpdate)
			RaidDebuffFrame_OnUpdate(self, 1)
		else
			self.timeText:SetText()
			self:SetScript("OnUpdate", nil)
		end

		self:Show()
	else
		self:Hide()
	end
end

local function UpdateCallback(overlay, unit)
	local frame = overlay.RaidDebuff
	if not unit or unit == frame.unit then
		RaidDebuffFrame_UpdateDebuff(frame)
		if unit then
			return 1 -- unit found, abort enumeration
		end
	end
end

function module:ApplyDebuffScale(overlay, scale)
	overlay.RaidDebuff:SetScale(scale)
end

function module:ApplyDebuffOffset(overlay, xoffset, yoffset)
	local frame = overlay.RaidDebuff
	frame:ClearAllPoints()
	frame:SetPoint("CENTER", frame.anchor, "CENTER", xoffset, yoffset)
end

function module:SetDebuffLevel(spellId, level)
	if type(spellId) ~= "number" then
		return
	end

	local name = GetSpellInfo(spellId)
	if not name then
		return
	end

	if type(level) == "number" then
		level = NormalizeLevel(level)
	else
		level = nil
	end

	local instance
	for _, instance in ipairs(self.instanceList) do
		local data = instance.debuffs[name]
		if data then
			data.level = level or data.defLevel
		end
	end
end

function module:OnInitialize(db, chardb, firstTime)
	db.scale = self.optionPage.scaleoffset:SetValue("scale", db.scale)
	db.xoffset, db.yoffset = self.optionPage.scaleoffset:SetValue("offset", db.xoffset, db.yoffset)
	CompactRaid:EnumUnitFrames(CreateDebuffFrame)

	if type(db.userData) ~= "table" then
		db.userData = {}
	end

	local instanceName, instanceData
	for instanceName, instanceData in pairs(db.userData) do
		if type(instanceName) == "string" and type(instanceData) == "table" and type(instanceData.bosses) == "table" then
			-- try register the instance
			local instance, _, new = self:RegisterInstance(instanceName)
			if instance then
				if new then
					instance.user = 1
				end

				local bossName, debuffs
				for bossName, debuffs in pairs(instanceData.bosses) do
					if type(bossName) == "string" and type(debuffs) == "table" then
						local boss, _, new = instance:RegisterBoss(bossName)
						if boss then
							if new then
								boss.user = 1
							end

							local id
							for id in pairs(debuffs) do
								local debuff = boss:RegisterDebuff(id)
								if debuff then
									debuff.user = 1
								else
									debuffs[id] = nil
								end
							end
						else
							instanceData.bosses[bossName] = nil
						end
					else
						-- invalid boss data, delete it
						instanceData.bosses[bossName] = nil
					end
				end
			else
				db.userData[instanceName] = nil
			end
		else
			-- invalid instance data, delete it
			db.userData[instanceName] = nil
		end
	end

	-- Apply user-modified priorities
	local id, level
	for id, level in pairs(chardb) do
		self:SetDebuffLevel(id, level)
	end

	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:ZONE_CHANGED_NEW_AREA()
end

function module:ZONE_CHANGED_NEW_AREA()
	--module:RegisterInstance(GetRealZoneText()):RegisterBoss():RegisterDebuff(6788) -- debug, weaken sould
	local instance = self:GetInstance(GetRealZoneText())
	activeDebuffs = instance and instance.debuffs
end

function module:OnUnitChange(overlay, unit, realUnit, inVehicle)
	local frame = overlay.RaidDebuff
	frame.unit = unit
	RaidDebuffFrame_UpdateDebuff(frame)
end

function module:OnShow(overlay)
	local frame = overlay.RaidDebuff
	frame:RegisterEvent("UNIT_AURA")
	RaidDebuffFrame_UpdateDebuff(frame)
end

function module:OnHide(overlay)
	overlay.RaidDebuff:UnregisterEvent("UNIT_AURA")
end

local function Frame_OnEvent(self, event, unit)
	if unit and self.unit and UnitIsUnit(unit, self.unit) then
		RaidDebuffFrame_UpdateDebuff(self)
	end
end

function module:OnCreate(overlay)
	-- Create raid debuff icon
	local frame = CreateFrame("Frame", nil, overlay)
	overlay.RaidDebuff = frame
	frame:SetSize(20, 20)

	frame.anchor = overlay:GetParent().healthBar
	frame:SetPoint("CENTER", frame.anchor, "CENTER")
	frame:Hide()

	frame.bkgnd = frame:CreateTexture(nil, "BORDER")
	frame.bkgnd:SetAllPoints(frame)

	frame.icon = frame:CreateTexture(nil, "ARTWORK")
	frame.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	frame.icon:SetPoint("TOPLEFT", 1, -1)
	frame.icon:SetPoint("BOTTOMRIGHT", -1, 1)

	frame.timeText = frame:CreateFontString(nil, "OVERLAY", "TextStatusBarText")
	frame.timeText:SetFont(STANDARD_TEXT_FONT, 10, "OUTLINE")
	frame.timeText:SetPoint("TOPLEFT", 1, -1)

	frame.countText = frame:CreateFontString(nil, "OVERLAY", "TextStatusBarText")
	frame.countText:SetFont(STANDARD_TEXT_FONT, 12, "OUTLINE")
	frame.countText:SetPoint("BOTTOMRIGHT", 0, 1)
	frame:SetScript("OnEvent", Frame_OnEvent)

	module:ApplyDebuffScale(overlay, (module.db.scale or 100) / 100)
	module:ApplyDebuffOffset(overlay, module.db.xoffset, module.db.yoffset)

	return frame
end
