------------------------------------------------------------
-- ModuleHandler.lua
--
-- Abin
-- 2012/1/03
------------------------------------------------------------

local _G = _G
local type = type
local GetLocale = GetLocale
local DEFAULT_CHAT_FRAME = DEFAULT_CHAT_FRAME
local tostring = tostring
local CreateFrame = CreateFrame
local format = format
local pairs = pairs
local StaticPopup_Show = StaticPopup_Show

local _, addon = ...

addon.guid = "{8A6673B0-DEA9-40DA-A955-FB60FBD0ABAC}"
addon.OBSOLETE_MODULES = { ["BarTextures"] = 1, ["RaidDebuff_Records"] = 1 } -- Obsolete modules, reject them from registering

local modules = {}
local moduleLocales = {}

local L = addon.L

function addon:RegisterLocale(name, locale, data)
	if type(name) == "string" and type(data) == "table" and (GetLocale() == locale or not moduleLocales[name]) then
		moduleLocales[name] = data
	end
end

function addon:GetLocale(name)
	return moduleLocales[name]
end

local function Print(module, msg, r, g, b)
	if msg then
		local header = addon.name
		if type(module) == "table" then
			header = header.."-"..(module.title or module.name or UNKNOWN)
		end
		DEFAULT_CHAT_FRAME:AddMessage("|cffffff78"..header..":|r "..tostring(msg), r or 0.5, g or 0.75, b or 1) -- All modules use uniformed message color, please!
	end
end

function addon:GetModule(name)
	return modules[name]
end

local function Module_RegisterEvent(module, event, method)
	if type(method) ~= "function" and type(method) ~= "string" then
		method = nil
	end

	module.__eventFrame.events[event] = method
	if not module.__eventFrame:IsEventRegistered(event) then
		module.__eventFrame:RegisterEvent(event)
	end
end

local function Module_UnregisterEvent(module, event)
	module.__eventFrame.events[event] = nil
	if module.__eventFrame:IsEventRegistered(event) then
		module.__eventFrame:UnregisterEvent(event)
	end
end

local function Module_IsEventRegistered(module, event)
	return module.__eventFrame:IsEventRegistered(event)
end

local function Module_RegisterAllEvents(module)
	return module.__eventFrame:RegisterAllEvents()
end

local function Module_UnregisterAllEvents(module)
	return module.__eventFrame:UnregisterAllEvents()
end

local function EventFrame_OnEvent(self, event, ...)
	local parent = self.parent
	if type(parent.OnEvent) == "function" then
		parent:OnEvent(event, ...)
	else
		local func = self.events[event]
		if not func then
			func = parent[event]
		elseif type(func) ~= "function" then -- string, number, etc
			func = parent[func]
		end

		if type(func) == "function" then
			func(parent, ...)
		end
	end
end

-- Register a new module, must be called before "ADDON_LOADED" event fires for addon named addonName
function addon:RegisterModule(addonName, name, title, version, desc, dualTalents, parent)
	if type(addonName) ~= "string" or type(name) ~= "string" or modules[name] or self.OBSOLETE_MODULES[name] then
		return
	end

	local module = {}
	module.addonName = addonName
	module.name = name
	module.title = title
	module.version = version
	module.desc = desc
	module.registeredFrames = {}
	module.dualTalentSupport = dualTalents
	module.Print = Print
	module.pcall = addon.pcall
	module.RequestRestoreDefaults = addon.RestoreModuleDefaults
	module.CombatSafeCall = addon.CombatSafeCall
	module.CancelPendingCombatSafeCall = addon.CancelPendingCombatSafeCall

	if title then
		if type(parent) == "table" then
			parent = parent.title
		end

		module.parent = type(parent) == "string" and parent or nil
		module.optionPage = self.optionTemplates:CreateOptionFrame(nil, title, version, desc, dualTalents, module.parent)
		module.optionPage.moduleName = name
	end

	local frame = CreateFrame("Frame")
	module.__eventFrame = frame
	frame.parent = module
	frame.events = {}
	frame:SetScript("OnEvent", EventFrame_OnEvent)
	module.RegisterEvent = Module_RegisterEvent
	module.UnregisterEvent = Module_UnregisterEvent
	module.IsEventRegistered = Module_IsEventRegistered
	module.RegisterAllEvents = Module_RegisterAllEvents
	module.UnregisterAllEvents = Module_UnregisterAllEvents

	modules[name] = module
	return module
end

local TALENT_SPEC_PRIMARY = TALENT_SPEC_PRIMARY
local TALENT_SPEC_SECONDARY = TALENT_SPEC_SECONDARY

local function NotifyTalentSwitch(module, spec)
	if not module.dualTalentSupport or not module._moduleInitDone then
		return
	end

	local talentName = spec == 2 and TALENT_SPEC_SECONDARY or TALENT_SPEC_PRIMARY
	local page = module.optionPage
	if page and page.dualTalentText then
		page.dualTalentText:SetText("["..talentName.."]")
	end

	local key = "talent"..spec
	local firstTime
	if type(module.chardb[key]) ~= "table" then
		module.chardb[key] = {}
		firstTime = 1
	end
	module.talentdb = module.chardb[key]

	if module:pcall(module.OnTalentGroupChange, module, spec, module.talentdb, firstTime) then
		module:Print(format(L["talent group message"], talentName))
	end
end

function addon:NotifyTalentSwitch()
	local spec = GetActiveTalentGroup() or 1
	local module
	for _, module in pairs(modules) do
		NotifyTalentSwitch(module, spec)
	end
end

function addon._EnumFrameProc_OnCreate(overlay, module)
	local frame = overlay:GetParent()
	addon:CallModuleFunc(module, "OnCreate", overlay)
	module.registeredFrames[frame] = 1
	if overlay:IsVisible() then
		addon:CallModuleFunc(module, "OnUnitChange", overlay, frame.displayedUnit, frame.unit, frame.inVehicle)
		addon:CallModuleFunc(module, "OnShow", overlay)
	end
end

local function EnumModuleProc_Initialize(name, module, addonName)
	if module._moduleInitDone or module.addonName ~= addonName then
		return
	end

	local firstTime
	local db = addon.db.modules[name]
	if type(db) ~= "table" then
		firstTime = 1
		db = {}
		addon.db.modules[name] = db
	end

	local chardb = addon.chardb.modules[name]
	if type(chardb) ~= "table" then
		firstTime = 1
		chardb = {}
		addon.chardb.modules[name] = chardb
	end

	module.db = db
	module.chardb = chardb
	addon.optionFrame:AddCategory(module.title, module.optionPage, module.parent)
	addon:CallModuleFunc(module, "OnInitialize", db, chardb, firstTime)
	addon:EnumUnitFrames(addon._EnumFrameProc_OnCreate, module)
	module._moduleInitDone = 1
	NotifyTalentSwitch(module, GetActiveTalentGroup() or 1)
end

function addon:ProcessModuleLoading(addonName)
	if type(addonName) == "string" and self:IsInitialized() then
		self:EnumModules(EnumModuleProc_Initialize, addonName)
	end
end

function addon:EnumModules(func, ...)
	if type(func) == "function" then
		local name, module
		for name, module in pairs(modules) do
			func(name, module, ...)
		end
	end
end

function addon:CallModuleFunc(module, func, ...)
	func = module and module[func]
	if type(func) == "function" then
		-- Why pcall? One module's error won't halt the entire process. I do need a way to let the module know
		-- it's encountering errors, though, so just print out the error message. You got a better idea?
		module:pcall(func, module, ...)
	end
end

function addon:CallAllModulesFunc(func, ...)
	local module
	for _, module in pairs(modules) do
		addon:CallModuleFunc(module, func, ...)
	end
end

function addon:NotifyFrameToAllModules(func, frame, ...)
	if not self:IsInitialized() then
		return
	end

	local module
	for _, module in pairs(modules) do
		if module._moduleInitDone then
			if func == "OnCreate" then
				if not module.registeredFrames[frame] then
					module.registeredFrames[frame] = 1
					addon:CallModuleFunc(module, func, frame.overlay, ...)
				end
			else
				if module.registeredFrames[frame] then
					addon:CallModuleFunc(module, func, frame.overlay, ...)
				end
			end
		end
	end
end

------------------------------------------------------------
-- For modules' "Restore Defaults" features, new feature added in 0.2 beta
------------------------------------------------------------

local popupData = {
	button1 = OKAY,
	button2 = CANCEL,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
	OnAccept = function(self, module, arg1)
		if type(module.OnRestoreDefaults) == "function" then
			if module:pcall(module.OnRestoreDefaults, module, arg1) then
				module:Print(text or L["defaults restored"])
			end
		end
	end,
}

StaticPopupDialogs[addon.guid] = popupData -- If you ever wondered why there's a guid for CompactRaid, here you go.

function addon:RestoreModuleDefaults(module, text, arg1)
	popupData.text = text or format(L["restore defaults text"], module.title or module.name or UNKNOWN)
	local dialog = StaticPopup_Show(addon.guid)
	if dialog then
		dialog.data = module
		dialog.data2 = arg1
		return dialog
	end
end