------------------------------------------------------------
-- UnitFrames.lua
--
-- Abin
-- 2012/1/03
------------------------------------------------------------

local _G = _G
local type = type
local ipairs = ipairs
local UnitIsConnected = UnitIsConnected
local UnitIsDead = UnitIsDead
local UnitIsPlayer = UnitIsPlayer
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local select = select
local UnitClass = UnitClass
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitPlayerControlled = UnitPlayerControlled
local UnitCanAttack = UnitCanAttack
local UnitReaction = UnitReaction
local DebuffTypeColor = DebuffTypeColor
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local wipe = wipe
local UnitGroupRolesAssigned = UnitGroupRolesAssigned
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitGetIncomingHeals = UnitGetIncomingHeals
local UnitHasVehicleUI = UnitHasVehicleUI
local strmatch = strmatch
local UnitPowerType = UnitPowerType
local PowerBarColor = PowerBarColor
local UnitPowerMax = UnitPowerMax
local UnitName = UnitName
local UnitIsUnit = UnitIsUnit
local UnitThreatSituation = UnitThreatSituation
local GetReadyCheckStatus = GetReadyCheckStatus
local GetTime = GetTime
local GetRaidTargetIndex = GetRaidTargetIndex
local SetRaidTargetIconTexture = SetRaidTargetIconTexture
local UnitHasIncomingResurrection = UnitHasIncomingResurrection
local UnitInRange = UnitInRange
local UnitIsVisible = UnitIsVisible
local InCombatLockdown = InCombatLockdown
local GameTooltip = GameTooltip
local CreateFrame = CreateFrame
local tinsert = tinsert
local RegisterStateDriver = RegisterStateDriver
local RegisterUnitWatch = RegisterUnitWatch
local strlen = string.utf8len or strlen
local strsub = string.utf8sub or strsub

local _, addon = ...
addon.defaultTexture = "Interface\\BUTTONS\\WHITE8X8.BLP"

local unitFrames = {} -- Stores all unit frames that were created by this addon

-- Enumerate all created overlay frames, your callback function returning non-nil value will abort the process
----------------------------------------------------------------
-- The way how this function works depends on arguments list you pass:

-- Arguments			-- Results (frame is the overlay frame that is passed to your callback function)
----------------------------------------------------------------
-- func, ...			-- func(frame, ...)		-- func is a valid function
-- object, "method", ...	-- object:method(frame, ...)	-- object is a table, "method" is a member function name of object
-- nil, "method", ...		-- frame:method(...)		-- first argument is nil, "method" is a member function name of the frame, such as "SetScale", "Show", "Hide", etc
----------------------------------------------------------------

function addon:EnumUnitFrames(object, func, ...)
	local enumType, memberFunc
	if type(object) == "function" then
		enumType = 1 -- Direct function call, object(frame, func, ...)
	elseif type(func) == "string" then
		if not object then
			memberFunc = testData.overlay[func]
			if type(memberFunc) == "function" then
				enumType = 2 -- Call frame method, frame:func(...)
			end

		elseif type(object) == "table" then
			memberFunc = object[func]
			if type(memberFunc) == "function" then
				enumType = 3 -- Call the object's member method, object:func(frame, ...)
			end
		end
	end

	local count = 0
	local data, abort, frameMethod
	for _, data in ipairs(unitFrames) do
		count = count + 1
		local frame = data.overlay
		if enumType == 1 then
			abort = object(frame, func, ...)
		elseif enumType == 2 then
			abort = memberFunc(frame, ...)
		elseif enumType == 3 then
			abort = memberFunc(object, frame, ...)
		end

		if abort then
			break
		end
	end
	return count, abort
end

local function GetUnitColor(unit, isName)
	if not unit then
		return 1, 1, 1
	end

	if not isName and not UnitIsConnected(unit) then
		return 0.5, 0.5, 0.5
	end

	local dead = UnitIsDead(unit)
	if UnitIsPlayer(unit) then
		local color = RAID_CLASS_COLORS[select(2, UnitClass(unit))]
		if color then
			return color.r, color.g, color.b
		else
			return 1, 1, 1
		end
	elseif not isName and (dead or (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit))) then
		return 0.5, 0.5, 0.5
	elseif UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			return 1, 0, 0
		elseif UnitCanAttack("player", unit) then
			return 1, 1, 0
		else
			return 0.5, 0.5, 1 -- friendly pet
		end
	else
		local reaction = UnitReaction(unit, "player")
		if not reaction then
			return 0.5, 0.5, 1
		elseif reaction < 4 then
			return 1, 0, 0
		elseif reaction > 4 then
			return 0, 1, 0
		else
			return 1, 1, 0
		end
	end
end

local function UpdateAuras(self, name, icon, count, duration, expires, dispelType)
	if name then
		self.icon:SetTexture(icon)
		if count and count > 1 then
			self.count:SetText(count)
			self.count:Show()
		else
			self.count:Hide()
		end
		if expires and expires ~= 0 then
			self.cooldown:SetCooldown(expires - duration, duration)
			self.cooldown:Show()
		else
			self.cooldown:Hide()
		end

		if dispelType then
			local color = DebuffTypeColor[dispelType] or DebuffTypeColor["none"]
			self.border:SetVertexColor(color.r, color.g, color.b)
		end

		self:Show()
	else
		self:Hide()
	end
end

local function UnitFrame_UpdateBuffs(self)
	if addon.db.hidebuffs then
		return
	end

	local unit = self.displayedUnit
	if not unit then return end
	local i
	for i = 1, 3 do
		local frame = self.buffFrames[i]
		local name, _, icon, count, _, duration, expires = UnitBuff(unit, i, "RAID|PLAYER")
		UpdateAuras(frame, name, icon, count, duration, expires)
	end
end

local function UnitFrame_UpdateDebuffs(self)
	if addon.db.hidedebuffs then
		return
	end

	local unit = self.displayedUnit
	if not unit then return end

	local filter = addon.db.onlyDispellable and "RAID" or ""
	local index = 0
	local i
	for i = 1, 40 do
		if index >= 3 then
			break
		end

		local name, _, icon, count, dispelType, duration, expires = UnitDebuff(unit, i, filter)
		if not name then
			break
		end

		if duration and duration < 600 then
			index = index + 1
			UpdateAuras(self.debuffFrames[index], name, icon, count, duration, expires, dispelType or "none")
		end
	end

	for i = index + 1, 3 do
		self.debuffFrames[i]:Hide()
	end
end

local DISPELABLES = { Magic = 1, Curse = 1, Disease = 1, Poison = 1 }

local function UnitFrame_UpdateDispels(self)
	if addon.db.hidedispels then
		return
	end

	local unit = self.displayedUnit
	if not unit then return end
	wipe(self.dispelable)
	local i
	for i = 1, 3 do
		local frame = self.dispelFrames[i]
		local name, _, _, _, dispelType = UnitDebuff(unit, i, "RAID")
		if name and DISPELABLES[dispelType] and not self.dispelable[dispelType] then
			self.dispelable[dispelType] = i
			frame.icon:SetTexture("Interface\\RaidFrame\\Raid-Icon-Debuff"..dispelType)
			frame:Show()
		else
			frame:Hide()
		end
	end
end

local function UnitFrame_UpdateAuras(self)
	UnitFrame_UpdateBuffs(self)
	UnitFrame_UpdateDebuffs(self)
	UnitFrame_UpdateDispels(self)
end

local function UnitFrame_UpdateRole(self)
	local unit = self.unit
	if not unit then return end
	local icon = self.roleIcon
	local show
	if self.inVehicle then
		icon:SetTexture("Interface\\Vehicles\\UI-Vehicles-Raid-Icon")
		icon:SetTexCoord(0, 1, 0, 1)
		show = 1
	else
		local role = UnitGroupRolesAssigned(unit)
		show = role == "TANK" or role == "HEALER" or role == "DAMAGER"
		if show then
			icon:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES")
			icon:SetTexCoord(GetTexCoordsForRoleSmallCircle(role))
		end
	end

	if show then
		icon:Show()
	else
		icon:Hide()
	end
end

local function UnitFrame_UpdateHealthPrediction(self)
	local incomingHeal = self.incomingHeal
	local unit = self.displayedUnit
	if not unit or UnitIsDeadOrGhost(unit) then
		incomingHeal:Hide()
		return
	end

	local width = 0
	local healthBar = self.healthBar
	local _, maxValue = healthBar:GetMinMaxValues()
	local value = healthBar:GetValue()
	local lacks = maxValue - value
	local incoming
	if lacks > 1 then
		incoming = UnitGetIncomingHeals(unit)
		if incoming and incoming > 0 then
			if incoming > lacks then
				incoming = lacks
			end
			width = incoming / maxValue * healthBar:GetWidth()
		end
	end

	if width > 1 then
		incomingHeal:SetWidth(width)
		incomingHeal:Show()
	else
		incomingHeal:Hide()
	end
end

local function UnitFrame_UpdateVehicleStatus(self)
	local unit = self.unit
	if not unit then return end
	if UnitHasVehicleUI(unit) then
		self.inVehicle = 1
		if unit == "player" then
			self.displayedUnit = "vehicle"
		else
			local prefix, id, suffix = strmatch(unit, "([^%d]+)([%d]*)(.*)")
			self.displayedUnit = prefix.."pet"..id..suffix
		end
	else
		self.inVehicle = nil
		self.displayedUnit = unit
	end
end

local function UnitFrame_UpdateHealthMax(self)
	local unit = self.displayedUnit
	if not unit then return end
	self.healthBar:SetMinMaxValues(0, UnitHealthMax(unit))
end

local function UnitFrame_UpdateHealth(self)
	local unit = self.displayedUnit
	if not unit then return end
	self.healthBar:SetValue(UnitHealth(unit))
end

local function UnitFrame_UpdateHealthColor(self)
	local unit = self.unit
	if not unit then return end
	local r, g, b
	if not UnitIsConnected(unit) then
		r, g, b = 0.5, 0.5, 0.5
	elseif addon.db.forceHealthColor then
		r, g, b = addon.db.healthr or 0, addon.db.healthg or 1, addon.db.healthb or 0
	elseif self.inVehicle then
		r, g, b = 0, 1, 0
	else
		r, g, b = GetUnitColor(unit)
	end

	self.incomingHeal:SetVertexColor(r * 0.5, g * 0.5, b * 0.5)
	local dr, dg, db = r * 0.25, g * 0.25, b * 0.25
	if addon.db.invertColor then
		self.healthBarBackground:SetVertexColor(dr, dg, db)
		self.healthBar:SetStatusBarColor(r, g, b)
	else
		self.healthBarBackground:SetVertexColor(r, g, b)
		self.healthBar:SetStatusBarColor(dr, dg, db)
	end
end

local function UnitFrame_UpdatePowerType(self)
	if not self.powerBar then return end
	local unit = self.displayedUnit
	if not unit then return end
	local r, g, b
	if not UnitIsConnected(unit) then
		r, g, b = 0.5, 0.5, 0.5
	elseif addon.db.forcePowerColor then
		r, g, b = addon.db.powerr or 0, addon.db.powerg or 0, addon.db.powerb or 1
	else
		local powerType, powerToken, altR, altG, altB = UnitPowerType(unit)
		local prefix = _G[powerToken]
		local info = PowerBarColor[powerToken]
		if info then
			r, g, b = info.r, info.g, info.b
		else
			if not altR then
				info = PowerBarColor[powerType] or PowerBarColor["MANA"]
				r, g, b = info.r, info.g, info.b
			else
				r, g, b = altR, altG, altB
			end
		end
	end
	local dr, dg, db = r * 0.25, g * 0.25, b * 0.25
	if addon.db.invertColor then
		self.powerBarBackground:SetVertexColor(dr, dg, db)
		self.powerBar:SetStatusBarColor(r, g, b)
	else
		self.powerBarBackground:SetVertexColor(r, g, b)
		self.powerBar:SetStatusBarColor(dr, dg, db)
	end
end

local function UnitFrame_UpdatePowerMax(self)
	if not self.powerBar then return end
	local unit = self.displayedUnit
	if not unit then return end
	local value = UnitPowerMax(unit)
	if value > 0 then
		self.powerBar:SetMinMaxValues(0, value)
		self.powerBar:Show()
		self.powerBarBackground:Show()
	else
		self.powerBar:Hide()
		self.powerBarBackground:Hide()
	end
end

local function UnitFrame_UpdatePower(self)
	if not self.powerBar then return end
	local unit = self.displayedUnit
	if not unit then return end
	self.powerBar:SetValue(UnitPower(unit))
end

local function UnitFrame_UpdateNameColor(self)
	local unit = self.unit
	if not unit then return end
	if addon.db.forceNameColor then
		self.nameText:SetTextColor(addon.db.namer or 1, addon.db.nameg or 1, addon.db.nameb or 1)
	else
		self.nameText:SetTextColor(GetUnitColor(unit, 1))
	end
end

local function UnitFrame_UpdateNameSize(self)
	local unit = self.unit
	if not unit then return end

	local text = UnitName(unit)
	local name = self.nameText
	if not text then
		name:SetText()
		return
	end

	local buttonWidth = self:GetWidth()
	if buttonWidth < 1 then
		return
	end

	local widthLimit = addon.db.nameWidthLimit
	if type(widthLimit) == "number" then
		widthLimit = widthLimit / 100
	else
		widthLimit = 0.75
	end

	local i
	for i = strlen(text), 0, -1 do
		name:SetText(strsub(text, 1, i))
		if name:GetWidth() / buttonWidth <= widthLimit then
			break
		end
	end
end

local function UnitFrame_UpdateName(self)
	UnitFrame_UpdateNameSize(self)
	UnitFrame_UpdateNameColor(self)
	addon:NotifyFrameToAllModules("OnUnitChange", self, self.displayedUnit, self.unit, self.inVehicle)
end

local function UnitFrame_UpdateStatusText(self)
	local unit = self.displayedUnit
	if not unit then return end

	local mode = addon.db.healthtextmode
	local text
	if not UnitIsConnected(unit) then
		text = PLAYER_OFFLINE
	elseif UnitIsDeadOrGhost(unit) then
		text = DEAD
	elseif mode == 3 then -- health percentage
		local healthMax = UnitHealthMax(unit)
		if healthMax > 0 then
			local health = UnitHealth(unit)
			if health < healthMax then
				text = format("%d%%", health / healthMax * 100)
			end
		end
	elseif mode == 2 then -- health lost
		local lost = UnitHealthMax(unit) - UnitHealth(unit)
		if lost > 0 then
			text = format("-%d", lost)
		end
	elseif mode == 1 then -- health remain
		text = format("%d", UnitHealth(unit))
	end
	self.statusText:SetText(text)
end

local function UnitFrame_UpdateTarget(self)
	local unit = self.displayedUnit
	if not unit then return end
	if UnitIsUnit(unit, "target") then
		self.selectionHighlight:Show()
	else
		self.selectionHighlight:Hide()
	end
end

local function UnitFrame_UpdateThreat(self)
	local unit = self.displayedUnit
	if not unit or not UnitExists(unit) then return end
	local status = UnitThreatSituation(unit)
	if status and status > 0 then
		self.aggroHighlight:SetVertexColor(GetThreatStatusColor(status))
		self.aggroHighlight:Show()
	else
		self.aggroHighlight:Hide()
	end
end

local READYCHECK_STATUS = { ready = READY_CHECK_READY_TEXTURE, notready = READY_CHECK_NOT_READY_TEXTURE, waiting = READY_CHECK_WAITING_TEXTURE }

local function UnitFrame_UpdateReadyCheck(self)
	local unit = self.unit
	if not unit then return end
	local status = GetReadyCheckStatus(unit)
	self.readyCheckStatus = status
	local texture = READYCHECK_STATUS[status]
	if texture then
		self.readyCheckIcon:SetTexture(texture)
		self.readyCheckIcon:Show()
	else
		self.readyCheckIcon:Hide()
	end
end

local function UnitFrame_FinishReadyCheck(self)
	if self.readyCheckStatus == "waiting" then
		self.readyCheckIcon:SetTexture(READY_CHECK_NOT_READY_TEXTURE)
	end
	self.readyCheckHideTime = GetTime() + 6
end

local function UnitFrame_DelayHideReadyCheck(self)
	local hideTime = self.readyCheckHideTime
	if hideTime and GetTime() > hideTime then
		self.readyCheckHideTime = nil
		self.readyCheckIcon:Hide()
	end
end

local function UnitFrame_UpdateRaidIcon(self)
	local unit = self.displayedUnit
	if not unit then return end
	local icon = self.raidIcon
	local index = GetRaidTargetIndex(unit)
	if index then
		SetRaidTargetIconTexture(icon, index)
		icon:Show()
	else
		icon:Hide()
	end
end

local function UnitFrame_UpdateResurrect(self)
	local unit = self.unit
	if not unit then return end
	if UnitHasIncomingResurrection(unit) then
		self.resIcon:Show()
		self.flagIcon:SetAlpha(0)
	else
		self.resIcon:Hide()
		self.flagIcon:SetAlpha(1)
	end
end

local function UnitFrame_UpdateFlags(self)
	local unit = self.unit
	if not unit then return end
	local icon = self.flagIcon
	local texture, flag
	if UnitIsDead(unit) then
		flag = "dead"
		texture = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull"
	elseif UnitIsGhost(unit) then
		flag = "ghost"
		texture = "Interface\\Icons\\Ability_Vanish"
	end

	if texture then
		icon:SetTexture(texture)
		icon:Show()
	else
		icon:Hide()
	end

	UnitFrame_UpdateStatusText(self)

	if flag ~= self.deathFlag then
		self.deathFlag = flag
		addon:NotifyFrameToAllModules("OnDeathFlagChange", self, flag)
	end
end

local function UnitFrame_UpdateInRange(self)
	local unit = self.displayedUnit
	if not unit then return end
	local inRange, checked = UnitInRange(unit)
	if not checked then
		inRange = UnitIsVisible(unit)
	end

	inRange = inRange and 1 or nil
	if inRange then
		self:SetAlpha(1)
	else
		self:SetAlpha(0.4)
	end

	if inRange ~= self.inRange then
		self.inRange = inRange
		addon:NotifyFrameToAllModules("OnRangeChange", self, inRange)
	end
end

local function UnitFrame_UpdateAll(self)
	UnitFrame_UpdateVehicleStatus(self)
	UnitFrame_UpdateResurrect(self)
	UnitFrame_UpdateFlags(self)
	UnitFrame_UpdateAuras(self)
	UnitFrame_UpdateRole(self)
	UnitFrame_UpdateHealthColor(self)
	UnitFrame_UpdateHealthMax(self)
	UnitFrame_UpdateHealth(self)
	UnitFrame_UpdateHealthPrediction(self)
	UnitFrame_UpdatePowerType(self)
	UnitFrame_UpdatePowerMax(self)
	UnitFrame_UpdatePower(self)
	UnitFrame_UpdateName(self)
	UnitFrame_UpdateTarget(self)
	UnitFrame_UpdateThreat(self)
	UnitFrame_UpdateReadyCheck(self)
	UnitFrame_UpdateInRange(self)
	UnitFrame_UpdateRaidIcon(self)
end

local function UnitFrame_RegisterEvents(self)
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_HEALTH")
	self:RegisterEvent("UNIT_MAXHEALTH")
	if self.powerBar then
		self:RegisterEvent("UNIT_POWER")
		self:RegisterEvent("UNIT_MAXPOWER")
		self:RegisterEvent("UNIT_DISPLAYPOWER")
		self:RegisterEvent("UNIT_POWER_BAR_SHOW")
		self:RegisterEvent("UNIT_POWER_BAR_HIDE")
	end
	self:RegisterEvent("UNIT_NAME_UPDATE")
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")
	self:RegisterEvent("UNIT_CONNECTION")
	self:RegisterEvent("UNIT_HEAL_PREDICTION")
	self:RegisterEvent("PLAYER_ROLES_ASSIGNED")
	self:RegisterEvent("UNIT_ENTERED_VEHICLE")
	self:RegisterEvent("UNIT_EXITED_VEHICLE")
	self:RegisterEvent("READY_CHECK")
	self:RegisterEvent("READY_CHECK_FINISHED")
	self:RegisterEvent("READY_CHECK_CONFIRM")
	self:RegisterEvent("PARTY_MEMBER_DISABLE")
	self:RegisterEvent("PARTY_MEMBER_ENABLE")
	self:RegisterEvent("UNIT_HEALTH_FREQUENT")
	self:RegisterEvent("INCOMING_RESURRECT_CHANGED")
	self:RegisterEvent("RAID_TARGET_UPDATE")
	self:RegisterEvent("UNIT_FLAGS")
	UnitFrame_UpdateAll(self)
end

local function UnitFrame_UnregisterEvents(self)
	self:UnregisterAllEvents()
end

local function UnitFrame_OnShow(self)
	UnitFrame_RegisterEvents(self)
	self.mayNeedUpdate = 1
	if self.affectContainer then
		addon:UpdateContainerSize()
	end
	addon:NotifyFrameToAllModules("OnShow", self)
end

local function UnitFrame_OnHide(self)
	UnitFrame_UnregisterEvents(self)
	if self.affectContainer then
		addon:UpdateContainerSize()
	end

	addon:NotifyFrameToAllModules("OnHide", self)
end

local function UnitFrame_OnEnter(self)
	local unit = self.displayedUnit
	if not unit then return end
	self.highlight:Show()

	if not (addon.db.showtooltip == 0 or (addon.db.showtooltip == 1 and InCombatLockdown())) then
		addon:SetTooltipAnchor()
		GameTooltip:SetUnit(unit)
	end

	addon:NotifyFrameToAllModules("OnEnter", self)
end

local function UnitFrame_OnLeave(self)
	GameTooltip:Hide()
	self.highlight:Hide()
	addon:NotifyFrameToAllModules("OnLeave", self)
end

local function UnitFrame_OnTick(self)
	UnitFrame_UpdateInRange(self)
	UnitFrame_DelayHideReadyCheck(self)
	if self.mayNeedUpdate then
		self.mayNeedUpdate = nil
		UnitFrame_UpdateAll(self)
	elseif self.flagsMayChanged then
		self.flagsMayChanged = nil
		UnitFrame_UpdateFlags(self)
	end
end

local function UnitFrame_OnUpdate(self, elapsed)
	local e = (self.__updateElapsed or 0) + elapsed
	if e > 0.2 then
		UnitFrame_OnTick(self)
		e = 0
	end
	self.__updateElapsed = e
end

local function UnitFrame_OnAttributeChanged(self, name, value)
	if name == "unit" then
		self.unit = value
		self.displayedUnit = value
	end
end

local function UnitFrame_ProcessUnitEvent(self, event, eventUnit)
	if not eventUnit then
		return
	end

	local unit, displayedUnit = self.unit, self.displayedUnit
	if not unit or not displayedUnit then
		return
	end

	if event == "UNIT_ENTERED_VEHICLE" or event == "UNIT_EXITED_VEHICLE" or event == "UNIT_CONNECTION" then
		if eventUnit == unit or UnitIsUnit(eventUnit, unit) then
			UnitFrame_UpdateAll(self)
		end
	elseif eventUnit == displayedUnit or UnitIsUnit(eventUnit, displayedUnit) then
		if event == "UNIT_MAXHEALTH" then
			UnitFrame_UpdateHealthMax(self)
			UnitFrame_UpdateHealth(self)
			UnitFrame_UpdateHealthPrediction(self)
			if (addon.db.healthtextmode or 0) > 0 then
				UnitFrame_UpdateStatusText(self)
			end
		elseif event == "UNIT_HEALTH" or event == "UNIT_HEALTH_FREQUENT" then
			UnitFrame_UpdateHealth(self)
			UnitFrame_UpdateHealthPrediction(self)
			if self.deathFlag then
				self.flagsMayChanged = 1
			elseif (addon.db.healthtextmode or 0) > 0 then
				UnitFrame_UpdateStatusText(self)
			end
		elseif event == "UNIT_HEAL_PREDICTION" then
			UnitFrame_UpdateHealthPrediction(self)
		elseif event == "UNIT_MAXPOWER" then
			UnitFrame_UpdatePowerMax(self)
			UnitFrame_UpdatePower(self)
		elseif event == "UNIT_POWER" then
			UnitFrame_UpdatePower(self)
		elseif event == "UNIT_DISPLAYPOWER" or event == "UNIT_POWER_BAR_SHOW" or event == "UNIT_POWER_BAR_HIDE" then
			UnitFrame_UpdatePowerMax(self)
			UnitFrame_UpdatePower(self)
			UnitFrame_UpdatePowerType(self)
		elseif event == "UNIT_AURA" then
			UnitFrame_UpdateAuras(self)
		elseif event == "UNIT_THREAT_SITUATION_UPDATE" then
			UnitFrame_UpdateThreat(self)
		elseif event == "READY_CHECK_CONFIRM" then
			UnitFrame_UpdateReadyCheck(self)
		elseif event == "INCOMING_RESURRECT_CHANGED" then
			UnitFrame_UpdateResurrect(self)
		elseif event == "UNIT_FLAGS" then
			self.flagsMayChanged = 1
		elseif event == "UNIT_NAME_UPDATE" then
			UnitFrame_UpdateAll(self)
		end
	end
end

local function UnitFrame_OnEvent(self, event, unit)
	if event == "PLAYER_ENTERING_WORLD" or event == "PARTY_MEMBERS_CHANGED" then
		UnitFrame_UpdateAll(self)
	elseif event == "PLAYER_TARGET_CHANGED" then
		UnitFrame_UpdateTarget(self)
	elseif event == "PLAYER_ROLES_ASSIGNED" then
		UnitFrame_UpdateRole(self)
	elseif event == "RAID_TARGET_UPDATE" then
		UnitFrame_UpdateRaidIcon(self)
	elseif event == "READY_CHECK" then
		UnitFrame_UpdateReadyCheck(self)
	elseif event == "READY_CHECK_FINISHED" then
		UnitFrame_FinishReadyCheck(self)
	elseif event == "PARTY_MEMBER_DISABLE" or event == "PARTY_MEMBER_ENABLE" then
		UnitFrame_UpdatePowerMax(self)
		UnitFrame_UpdatePower(self)
		UnitFrame_UpdatePowerType(self)
	else
		UnitFrame_ProcessUnitEvent(self, event, unit)
	end
end

local function CreateAuraFrame(key, name, parent, list)
	local frame = CreateFrame("Frame", name, parent, "CompactAuraTemplate")
	frame:Hide()
	frame:SetSize(11, 11)
	frame.key = key
	if key == "debuff" then
		local border = frame:CreateTexture(name.."Border", "OVERLAY")
		frame.border = border
		border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays")
		border:SetTexCoord(0.296875, 0.5703125, 0, 0.515625)
		border:SetAllPoints(frame)
	elseif key == "dispel" then
		frame.icon:SetTexCoord(0.125, 0.875, 0.125, 0.875)
	end
	tinsert(list, frame)
	return frame
end

local function InitUnitFrame(frame, visibility, mini)
	if frame.healthBar then
		return
	end

	frame:SetSize(64, mini and 18 or 36)
	frame:RegisterForClicks("AnyUp")
	frame:SetAttribute("type", "target")
	frame:SetAttribute("toggleForVehicle", true)
	frame:SetAttribute("ministyle", mini and 1 or nil)

	local name = frame:GetName()

	-- Background (black)
	local background = frame:CreateTexture(name.."Background", "BACKGROUND")
	frame.background = background
	background:SetAllPoints(frame)
	background:SetTexture(0, 0, 0, 1)

	-- Health bar
	local healthBar = CreateFrame("StatusBar", name.."HealthBar", frame)
	frame.healthBar = healthBar
	healthBar:SetPoint("TOPLEFT", 1, -1)
	if mini then
		healthBar:SetPoint("BOTTOMRIGHT", -1, 1)
	else
		healthBar:SetPoint("BOTTOMRIGHT", -1, 2)
	end
	healthBar:SetStatusBarTexture(addon.defaultTexture)
	background = frame:CreateTexture(name.."HealthBarBackground", "BORDER")
	frame.healthBarBackground = background
	background:SetAllPoints(healthBar)
	background:SetTexture(addon.defaultTexture)

	if not mini then
		-- Power bar
		local powerBar = CreateFrame("StatusBar", name.."PowerBar", frame)
		frame.powerBar = powerBar
		powerBar:SetPoint("TOPLEFT", healthBar, "BOTTOMLEFT")
		powerBar:SetPoint("BOTTOMRIGHT", -1, 1)
		powerBar:SetHeight(5)
		powerBar:SetStatusBarTexture(addon.defaultTexture)
		background = frame:CreateTexture(name.."PowerBarBackground", "BORDER")
		frame.powerBarBackground = background
		background:SetAllPoints(powerBar)
		background:SetTexture(addon.defaultTexture)
	end

	-- Incoming healing bar
	local incomingHeal = healthBar:CreateTexture(name.."IncomingHealBar", "ARTWORK")
	frame.incomingHeal = incomingHeal
	incomingHeal:SetPoint("TOPLEFT", healthBar:GetStatusBarTexture(), "TOPRIGHT")
	incomingHeal:SetPoint("BOTTOMLEFT", healthBar:GetStatusBarTexture(), "BOTTOMRIGHT")
	incomingHeal:SetTexture(addon.defaultTexture)
	incomingHeal:Hide()

	-- Art frame
	local artFrame = CreateFrame("Frame", name.."ArtFrame", frame)
	frame.artFrame = artFrame
	artFrame:SetAllPoints(frame)
	artFrame:SetFrameLevel(60)

	-- Highlight texture
	local highlight = artFrame:CreateTexture(name.."Highlight", "BACKGROUND")
	frame.highlight = highlight
	highlight:SetAllPoints(frame)
	highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
	highlight:SetBlendMode("ADD")
	highlight:SetVertexColor(1, 1, 1, 0.4)
	highlight:Hide()

	-- Name text
	local nameText = artFrame:CreateFontString(name.."NameText", "BORDER", "GameFontHighlightSmall")
	frame.nameText = nameText
	nameText:SetPoint("CENTER", healthBar, "CENTER")
	nameText:SetFont(STANDARD_TEXT_FONT, 12)

	-- Status text
	local statusText = artFrame:CreateFontString(name.."StatusText", "BORDER", "GameFontDisableSmall")
	frame.statusText = statusText
	statusText:SetPoint("TOP", nameText, "BOTTOM", 0, 1)
	statusText:SetFont(STANDARD_TEXT_FONT, 10)

	-- Role icon
	local roleIcon = artFrame:CreateTexture(name.."RoleIcon", "ARTWORK")
	frame.roleIcon = roleIcon
	roleIcon:SetPoint("TOPLEFT")
	roleIcon:SetSize(12, 12)

	-- Flag icon (dead/ghost)
	local flagIcon = artFrame:CreateTexture(name.."FlagIcon", "ARTWORK")
	frame.flagIcon = flagIcon
	flagIcon:SetSize(18, 18)
	flagIcon:SetPoint("CENTER")
	flagIcon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

	-- Resurrect icon
	local resIcon = artFrame:CreateTexture(name.."ResIcon", "ARTWORK")
	frame.resIcon = resIcon
	resIcon:SetSize(18, 18)
	resIcon:SetPoint("CENTER")
	resIcon:SetTexture("Interface\\RaidFrame\\Raid-Icon-Rez")
	resIcon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	resIcon:Hide()

	-- Raid icon
	local raidIcon = artFrame:CreateTexture(name.."RaidIcon", "OVERLAY")
	frame.raidIcon = raidIcon
	raidIcon:SetSize(14, 14)
	raidIcon:SetPoint("TOP", 0, 3)
	raidIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	raidIcon:Hide()

	-- Ready check icon
	local readyCheckIcon = artFrame:CreateTexture(name.."ReadyCheckIcon", "OVERLAY")
	frame.readyCheckIcon = readyCheckIcon
	readyCheckIcon:SetSize(16, 16)
	readyCheckIcon:SetPoint("CENTER", healthBar, "CENTER")
	readyCheckIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	readyCheckIcon:Hide()

	-- Aggro highlight
	local aggroHighlight = artFrame:CreateTexture(name.."AggroHighlight", "BORDER")
	frame.aggroHighlight = aggroHighlight
	aggroHighlight:SetAllPoints(frame)
	aggroHighlight:SetTexture("Interface\\RaidFrame\\Raid-FrameHighlights")
	aggroHighlight:SetTexCoord(0.00781250, 0.55468750, 0.00781250, 0.27343750)
	aggroHighlight:Hide()

	-- Selection highlight
	local selectionHighlight = artFrame:CreateTexture(name.."SelectionHighlight", "ARTWORK")
	frame.selectionHighlight = selectionHighlight
	selectionHighlight:SetAllPoints(frame)
	selectionHighlight:SetTexture("Interface\\RaidFrame\\Raid-FrameHighlights")
	selectionHighlight:SetTexCoord(0.00781250, 0.55468750, 0.28906250, 0.55468750)
	selectionHighlight:Hide()

	-- Aura frames
	local buffParent = CreateFrame("Frame", name.."BuffParent", artFrame)
	local debuffParent = CreateFrame("Frame", name.."DeBuffParent", artFrame)
	local dispelParent = CreateFrame("Frame", name.."DispelParent", artFrame)
	frame.buffParent = buffParent
	frame.debuffParent = debuffParent
	frame.dispelParent = dispelParent
	frame.buffFrames = {}
	frame.debuffFrames = {}
	frame.dispelFrames = {}
	frame.dispelable = {}
	local i
	for i = 1, 3 do
		local buff = CreateAuraFrame("buff", name.."Buff"..i, buffParent, frame.buffFrames)
		if i == 1 then
			buff:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", -2, 1)
		else
			buff:SetPoint("RIGHT", frame.buffFrames[i - 1], "LEFT")
		end

		local debuff = CreateAuraFrame("debuff", name.."Debuff"..i, debuffParent, frame.debuffFrames)
		if i == 1 then
			debuff:SetPoint("BOTTOMLEFT", healthBar, "BOTTOMLEFT", 2, 1)
		else
			debuff:SetPoint("LEFT", frame.debuffFrames[i - 1], "RIGHT")
		end

		local dispel = CreateAuraFrame("dispel", name.."Dispel"..i, dispelParent, frame.dispelFrames)
		if i == 1 then
			dispel:SetPoint("TOPRIGHT", healthBar, "TOPRIGHT", -2, -1)
		else
			dispel:SetPoint("RIGHT", frame.dispelFrames[i - 1], "LEFT")
		end
	end

	-- Create an overlay frame for modules so they can place their own stuff on it without worrying about being obstructed
	local overlay = CreateFrame("Frame", name.."ModuleOverlayFrame", frame)
	frame.overlay = overlay
	overlay:SetAllPoints(frame)
	overlay:SetFrameLevel(100)

	-- Basic scripts
	frame:HookScript("OnAttributeChanged", UnitFrame_OnAttributeChanged)
	frame:HookScript("OnShow", UnitFrame_OnShow)
	frame:HookScript("OnHide", UnitFrame_OnHide)
	frame:HookScript("OnEnter", UnitFrame_OnEnter)
	frame:HookScript("OnLeave", UnitFrame_OnLeave)
	frame:SetScript("OnUpdate", UnitFrame_OnUpdate)
	frame:SetScript("OnEvent", UnitFrame_OnEvent)

	frame.UpdateBuffs = UnitFrame_UpdateBuffs
	frame.UpdateDebuffs = UnitFrame_UpdateDebuffs
	frame.UpdateDispels = UnitFrame_UpdateDispels
	frame.UpdateRole = UnitFrame_UpdateRole
	frame.UpdateHealthPrediction = UnitFrame_UpdateHealthPrediction
	frame.UpdateVehicleStatus = UnitFrame_UpdateVehicleStatus
	frame.UpdateHealthMax = UnitFrame_UpdateHealthMax
	frame.UpdateHealth = UnitFrame_UpdateHealth
	frame.UpdateHealthColor = UnitFrame_UpdateHealthColor
	frame.UpdatePowerType = UnitFrame_UpdatePowerType
	frame.UpdatePowerMax = UnitFrame_UpdatePowerMax
	frame.UpdatePower = UnitFrame_UpdatePower
	frame.UpdateNameColor = UnitFrame_UpdateNameColor
	frame.UpdateName = UnitFrame_UpdateName
	frame.UpdateStatusText = UnitFrame_UpdateStatusText
	frame.UpdateTarget = UnitFrame_UpdateTarget
	frame.UpdateThreat = UnitFrame_UpdateThreat
	frame.UpdateReadyCheck = UnitFrame_UpdateReadyCheck
	frame.UpdateRaidIcon = UnitFrame_UpdateRaidIcon
	frame.UpdateResurrect = UnitFrame_UpdateResurrect
	frame.UpdateFlags = UnitFrame_UpdateFlags
	frame.UpdateNameSize = UnitFrame_UpdateNameSize
	frame.UpdateAuras = UnitFrame_UpdateAuras
	frame.UpdateAll = UnitFrame_UpdateAll

	tinsert(unitFrames, frame)
	if visibility ~= false then
		frame:Hide()
		if type(visibility) == "string" then
			RegisterStateDriver(frame, "visibility", visibility)
		else
			RegisterUnitWatch(frame)
		end
	end
end
addon.__InitUnitFrame = InitUnitFrame

local function UpdateUnitFrameSize(frame)
	if frame.powerBar then
		frame:SetSize(addon.db.width, addon.db.height)
	else
		frame:SetSize(addon.db.width, addon.db.height / 2)
	end
	UnitFrame_UpdateNameSize(frame)
end

local function UpdatePowerBarHeight(frame)
	if frame.powerBar then
		local healthBar = frame.healthBar
		healthBar:ClearAllPoints()
		healthBar:SetPoint("TOPLEFT", 1, -1)
		healthBar:SetPoint("BOTTOMRIGHT", -1, addon.db.powerBarHeight + 1)
	end
end

local function UpdateNameHeight(frame, font)
	font = font or STANDARD_TEXT_FONT
	local height = addon.db.nameHeight
	frame.nameText:SetFont(font, height)
	frame.statusText:SetFont(font, height * 0.84)
	UnitFrame_UpdateNameSize(frame)
end

local function UpdateNameOffsets(frame)
	local name = frame.nameText
	name:ClearAllPoints()
	name:SetPoint("CENTER", frame.healthBar, "CENTER", addon.db.nameXOffset, addon.db.nameYOffset)
end

local function UpdateAuraMode(frame)
	if addon.db.hideBuffs then
		frame.buffParent:Hide()
	else
		frame.buffParent:Show()
		UnitFrame_UpdateBuffs(frame)
	end

	if addon.db.hideDebuffs then
		frame.debuffParent:Hide()
	else
		frame.debuffParent:Show()
		UnitFrame_UpdateDebuffs(frame)
	end

	if addon.db.hideDispels then
		frame.dispelParent:Hide()
	else
		frame.dispelParent:Show()
		UnitFrame_UpdateDispels(frame)
	end
end

local function UpdateStatusBarBkgnd(frame)
	if addon.db.hidebarbkgnd then
		frame.healthBarBackground:Hide()
		if frame.powerBarBackground then
			frame.powerBarBackground:Hide()
		end
	else
		frame.healthBarBackground:Show()
		if frame.powerBarBackground then
			frame.powerBarBackground:Show()
		end
	end
end

function addon:GetDefaultEventProc()
	return UnitFrame_OnEvent
end

local DEFAULT_TEMPLATES = "SecureUnitButtonTemplate,SecureHandlerEnterLeaveTemplate" -- "SecureHandlerEnterLeaveTemplate" is specifically for ClickSet mouse-wheel supports

function addon:GetTemplates(additionalTemplates)
	if type(additionalTemplates) == "string" and additionalTemplates ~= "" then
		return DEFAULT_TEMPLATES..","..additionalTemplates
	else
		return DEFAULT_TEMPLATES
	end
end

function addon:CreateUnitFrame(name, parent, unit, visibility, mini, additionalTemplates)
	if type(name) ~= "string" then
		return
	end

	local frame = CreateFrame("Button", name, parent, self:GetTemplates(additionalTemplates))
	InitUnitFrame(frame, visibility, mini)
	if unit then
		frame:SetAttribute("unit", unit)
	end

	-- If the frame is created after ADDON_LOADED by some other addons
	if self:IsInitialized() then
		UpdateUnitFrameSize(frame)
		UpdateStatusBarBkgnd(frame)
		UpdatePowerBarHeight(frame)
		UpdateNameHeight(frame)
		UpdateNameOffsets(frame)
		UpdateAuraMode(frame)
		addon:CliqueRegister(frame)

		addon:NotifyFrameToAllModules("OnCreate", frame)
		if frame:IsVisible() then
			addon:NotifyFrameToAllModules("OnUnitChange", frame, frame.displayedUnit, frame.unit, frame.inVehicle)
			addon:NotifyFrameToAllModules("OnShow", frame)
		end
	end

	return frame
end

local function InernalEnum(func, ...)
	local frame
	for _, frame in ipairs(unitFrames) do
		func(frame, ...)
	end
end

function addon:UpdateUnitFrameSize()
	InernalEnum(UpdateUnitFrameSize)
	addon:UpdateRaidGroupHeaderSize()
	addon:UpdateContainerSize()
end

function addon:UpdatePowerBarHeight()
	InernalEnum(UpdatePowerBarHeight)
end

function addon:UpdateNameWidth()
	InernalEnum(UnitFrame_UpdateName)
end

function addon:UpdateNameHeight(font)
	InernalEnum(UpdateNameHeight, font)
end

function addon:UpdateNameColorMode()
	InernalEnum(UnitFrame_UpdateNameColor)
end

function addon:UpdateNameOffsets()
	InernalEnum(UpdateNameOffsets)
end

function addon:UpdateHealthColorMode()
	InernalEnum(UnitFrame_UpdateHealthColor)
end

function addon:UpdatePowerColorMode()
	InernalEnum(UnitFrame_UpdatePowerType)
end

function addon:UpdateAuraMode()
	InernalEnum(UpdateAuraMode)
end

function addon:UpdateStatusBarBkgnd()
	InernalEnum(UpdateStatusBarBkgnd)
end

function addon:UpdateHealthTextMode()
	InernalEnum(UnitFrame_UpdateStatusText)
end

-----------------------------------------------------
-- Clique supports
-----------------------------------------------------

local Clique = IsAddOnLoaded("Clique") and type(Clique) == "table" and type(Clique.RegisterFrame) == "function" and type(Clique.UnregisterFrame) == "function" and Clique

function addon:CliqueRegister(frame)
	if not Clique or not self:IsInitialized() then
		return
	end

	if frame then
		Clique:RegisterFrame(frame)
	else
		for _, frame in ipairs(unitFrames) do
			Clique:RegisterFrame(frame)
		end
	end
end

function addon:CliqueUnregister(frame)
	if not Clique or not self:IsInitialized() then
		return
	end

	if frame then
		Clique:UnregisterFrame(frame)
	else
		for _, frame in ipairs(unitFrames) do
			Clique:UnregisterFrame(frame)
		end
	end
end